////////////////////////////////////////////////////////////////////// 
// metal/Main.cpp 
// (c) 2000-2008 Goncalo Abecasis
// 
// This file is distributed as part of the Goncalo source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile Goncalo.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Friday April 18, 2008
// 
 
#include "StringArray.h"
#include "StringHash.h"
#include "MathVector.h"
#include "MathStats.h"
#include "IntArray.h"

#include <math.h>
#include <ctype.h>
#include <signal.h>

StringIntHash markerLookup;

Vector statistics;
Vector weights;
Vector pvalues;
Vector frequencies;
Vector frequencies2;
Vector minFrequencies;
Vector maxFrequencies;

StringArray allele1;
StringArray allele2;
String      original_flipped;

StringArray filenames;
StringArray directions;

double weight = 1.0;
double minweight = 1.0;

String markerLabel  = "MARKER";
String weightLabel  = "N";
String pvalueLabel  = "PVALUE";
String effectLabel  = "EFFECT";
String stderrLabel  = "STDERR";
String strandLabel  = "STRAND";
String frequencyLabel = "FREQ";
String firstAllele  = "ALLELE1";
String secondAllele = "ALLELE2";

String separators  = " \t";

bool   useStandardErrors = false;
bool   useStrand = false;
bool   averageFrequencies = false;
bool   minMaxFrequencies = false;
bool   genomicControl = false;

double lambda = 1.0;

int    maxWarnings = 20;

StringArray filterLabel;
IntArray    filterColumn;
IntArray    filterCondition;
Vector      filterValue;
StringArray filterAlternate;
IntArray    filterCounts;

void UserBreak(int)
   {
   printf("\n\n## METAL STOPPED BY USER\n\n");

   exit(EXIT_FAILURE);
   }

void OutOfMemory(int)
   {
   printf("\n\nMETAL HAS CRASHED\n\n"
          "The operating system has decided to terminate METAL.\n"
          "This could be due to a memory access problem, or it\n"
          "could be you encountered a bug.\n\n"
          "To help improve this program, please report bugs by\n"
          "e-mailing goncalo@umich.edu. Please include details\n"
          "of how the problem you encountered can be repdcued:);\n");

   exit(EXIT_FAILURE);
   }

typedef void (*signal_handler)(int);

void SetupCrashHandlers()
   {
   signal(SIGINT, (signal_handler) UserBreak);
   signal(SIGSEGV, (signal_handler) OutOfMemory);
   signal(SIGABRT, (signal_handler) OutOfMemory);
   }

void ClearAll()
   {
   printf("## Clearing all stored statistics %s...\n", filterColumn.Length() ? "and filters " : "");
   markerLookup.Clear();
   statistics.Dimension(0);
   weights.Dimension(0);
   pvalues.Dimension(0);
   allele1.Dimension(0);
   allele2.Dimension(0);
   frequencies.Dimension(0);
   frequencies2.Dimension(0);
   minFrequencies.Dimension(0);
   maxFrequencies.Dimension(0);
   original_flipped.Clear();
   filenames.Clear();
   directions.Clear();
   filterLabel.Dimension(0);
   filterCondition.Dimension(0);
   filterValue.Dimension(0);
   filterAlternate.Dimension(0);
   }

#define LESS_THAN              1
#define LESS_THAN_OR_EQUAL     2
#define EQUAL_TO               3
#define GREATER_THAN_OR_EQUAL  4
#define GREATER_THAN           5
#define NOT_EQUAL_TO           6
#define STRING_MATCH           (EQUAL_TO | 0x80)
#define STRING_MISMATCH        (NOT_EQUAL_TO | 0x80)

int TranslateCondition(String & condition)
   {
   if (condition == ">")
      return GREATER_THAN;
   else if (condition == ">=")
      return GREATER_THAN_OR_EQUAL;
   else if (condition == "<")
      return LESS_THAN;
   else if (condition == "<=")
      return LESS_THAN_OR_EQUAL;
   else if (condition == "==" || condition == "=" || condition == "IS")
      return EQUAL_TO;
   else if (condition == "!=" || condition == "<>" || condition == "ISNOT" )
      return NOT_EQUAL_TO;
   else
      return -1;
   }

void AddFilter(StringArray & filter)
   {
   if (filter.Length() != 4)
      {
      printf("# Filter ignored: Command requires exactly 3 arguments separated by whitespace\n"
             "#    The arguments correspond to a column label, condition, and filter value\n");
      return;
      }

   int condition = TranslateCondition(filter[2]);

   if (condition == -1)
      {
      printf("# Filter ignored: the condition '%s' is not recognized\n", (const char *) filter[2]);
      return;
      }

   bool isNumber = isdigit(filter[3][0]);

   if (!isNumber && condition != EQUAL_TO && condition != NOT_EQUAL_TO)
      {
      printf("# Filter ignored: the condition '%s' requires a numeric right-hand side value", (const char *) filter[2]);
      return;
      }

   if (!isNumber && condition == EQUAL_TO) condition = STRING_MATCH;
   if (!isNumber && condition == NOT_EQUAL_TO) condition = STRING_MISMATCH;

   filterLabel.Push(filter[1]);
   filterCondition.Push(condition);
   filterValue.Push(filter[3]);
   filterAlternate.Push(filter[3]);

   printf("# Added filter based on '%s' column\n", (const char *) filter[1]);
   }

void SetupFilters(StringArray & header)
   {
   filterColumn.Dimension(filterLabel.Length());
   filterCounts.Dimension(filterLabel.Length());
   filterCounts.Zero();

   for (int i = 0; i < filterLabel.Length(); i++)
      {
      filterColumn[i] = header.SlowFind(filterLabel[i]);

      if (filterColumn[i] < 0)
         printf("## WARNING: No '%s' column found -- filter based on this column will be ignored\n", (const char *) filterLabel[i]);
      }
   }

bool ApplyFilter(StringArray & row)
   {
   for (int i = 0; i < filterColumn.Length(); i++)
      if (filterColumn[i] >= 0)
         {
         String & text = row[filterColumn[i]];

         text.Trim();

         bool isValue = isdigit(text[0]) || text[0] == '.' && isdigit(text[1]);

         switch(filterCondition[i])
            {
            case LESS_THAN :
               if (!(row[filterColumn[i]].AsDouble() < filterValue[i]) || !isValue)
                  {
                  filterCounts[i]++;
                  return false;
                  }
               break;
            case LESS_THAN_OR_EQUAL :
               if (!(row[filterColumn[i]].AsDouble() <= filterValue[i]) || !isValue)
                  {
                  filterCounts[i]++;
                  return false;
                  }
               break;
            case GREATER_THAN :
               if (!(row[filterColumn[i]].AsDouble() > filterValue[i]) || !isValue)
                  {
                  filterCounts[i]++;
                  return false;
                  }
               break;
            case GREATER_THAN_OR_EQUAL :
               if (!(row[filterColumn[i]].AsDouble() >= filterValue[i]) || !isValue)
                  {
                  filterCounts[i]++;
                  return false;
                  }
               break;
            case EQUAL_TO :
               if (!(row[filterColumn[i]].AsDouble() == filterValue[i]) || !isValue)
                  {
                  filterCounts[i]++;
                  return false;
                  }
               break;
            case NOT_EQUAL_TO :
               if (!(row[filterColumn[i]].AsDouble() != filterValue[i]) || !isValue)
                  {
                  filterCounts[i]++;
                  return false;
                  }
               break;
            case STRING_MATCH :
               if (!(row[filterColumn[i]] == filterAlternate[i]))
                  {
                  filterCounts[i]++;
                  return false;
                  }
               break;
            case STRING_MISMATCH :
               if (!(row[filterColumn[i]] != filterAlternate[i]))
                  {
                  filterCounts[i]++;
                  return false;
                  }
               break;
            }
         }

   return true;
   }

void FilterSummary()
   {
   if (filterColumn.Length() == 0) return;

   if (filterCounts.Sum() == 0)
      {
      printf("## Although a filter is defined, no lines were excluded because of it\n");
      return;
      }

   const char * conditions[6] = {"<", "<=", "==", ">=", ">", "!="};

   printf("# %8d lines filtered, for the following reasons:\n", filterCounts.Sum());

   for (int i = 0; i < filterColumn.Length(); i++)
      if (filterCounts[i])
         printf("# %8d filtered due to constraint %s %s %s\n",
                filterCounts[i],
                (const char *) filterLabel[i],
                (const char *) conditions[(filterCondition[i] & 0x7F) - 1],
                (const char *) filterAlternate[i]);
   }

void ClearFilters()
   {
   printf("# Cleared user defined filters ...\n");
   filterLabel.Dimension(0);
   filterCondition.Dimension(0);
   filterValue.Dimension(0);
   filterAlternate.Dimension(0);
   filterCounts.Dimension(0);
   }

void NumbersToLetters(String & al)
   {
   if (al == "1") al = "a"; else
   if (al == "2") al = "c"; else
   if (al == "3") al = "g"; else
   if (al == "4") al = "t";
   }

void FlipAllele(String & al)
   {
   if (al == "A" || al == "a") al = "t"; else
   if (al == "C" || al == "c") al = "g"; else
   if (al == "G" || al == "g") al = "c"; else
   if (al == "T" || al == "t") al = "a";
   }

bool FlipAlleles(String & al1, String & al2, double & effect, double & freq)
   {
   al1.ToLower();
   al2.ToLower();

   if (al1 > al2)
      {
      al1.Swap(al2);
      effect *= -1.0;
      freq = 1.0 - freq;
      }

   if (al1 == "a" || al1 == "c" && al2 == "g")
      return false;

   FlipAllele(al1);
   FlipAllele(al2);

   if (al1 > al2)
      {
      al1.Swap(al2);
      effect *= -1.0;
      freq = 1.0 - freq;
      }

   return true;
   }

bool GuessSecondAllele(int marker, String & al1, String & al2)
   {
   al1.ToLower();

   FlipAllele(al2 = al1);

   if (allele1[marker] == al1 && allele2[marker] == al2 ||
       allele1[marker] == al2 && allele2[marker] == al1)
      return false;

   if (allele1[marker] == al1)
      {
      al2 = allele2[marker];
      return true;
      }
   else if (allele2[marker] == al1)
      {
      al2 = allele1[marker];
      return true;
      }
   else if (allele1[marker] == al2)
      {
      FlipAllele(al2 = allele2[marker]);
      return true;
      }
   else if (allele2[marker] == al2)
      {
      FlipAllele(al2 = allele1[marker]);
      return true;
      }

   return false;
   }

int    outfileround = 1;
String outfile("METAANALYSIS%d.TBL");

void Analyze()
   {
   String filename;
   filename.printf(outfile, outfileround++);

   printf("## Executing meta-analysis ...\n");
   printf("## Complete results will be stored in file '%s'\n", (const char *) filename);
   printf("## Column descriptions will be stored in file '%s.info'\n", (const char *) filename);

   FILE * f = fopen(filename, "wt");

   fprintf(f, "MarkerName\tAllele1\tAllele2\t%s%s%s\t%s\tP-value\tDirection\n",
              averageFrequencies ? "Freq1\tFreqSE\t" : "",
              minMaxFrequencies ? "MinFreq\tMaxFreq\t" : "",
              useStandardErrors ? "Effect" : "Weight",
              useStandardErrors ? "StdErr" : "Zscore");

   double smallp = 1.0;
   String smallm("nothing tested");

   String al1, al2;

   String direction;

   int count = 0, skipped = 0;
   for (int i = 0; i < markerLookup.Capacity(); i++)
      if (markerLookup.SlotInUse(i))
         {
         int marker = markerLookup.Integer(i);

         if (weights[marker] == 0.0) continue;

         if (!useStandardErrors && weights[marker] < minweight)
            {
            skipped++;
            continue;
            }

         double statistic = statistics[marker] / sqrt(weights[marker]);
         double frequency = averageFrequencies ? frequencies[marker] / weights[marker] : 0.5;
         double frequency2 = averageFrequencies ? frequencies2[marker] / weights[marker] : 0.5;
         double minFrequency = minMaxFrequencies ? minFrequencies[marker] : 0.5;
         double maxFrequency = minMaxFrequencies ? maxFrequencies[marker] : 0.5;
         double pvalue = 2.0 * ndist(fabs(statistic), true);

         frequency2 = frequency2 - frequency * frequency;
         frequency2 = frequency2 > 0.0 ? sqrt(frequency2) : 0.0;

         al1 = allele1[marker];
         al2 = allele2[marker];

         if (original_flipped[marker] == 'Y')
            FlipAllele(al1), FlipAllele(al2);

         direction.Clear();
         for (int j = 0; j < filenames.Length(); j++)
            direction += marker < directions[j].Length() ? directions[j][marker] : '?';

         fprintf(f, "%s\t%s\t%s\t",
                  (const char *) markerLookup[i],
                  (const char *) al1,
                  (const char *) al2);

         if (averageFrequencies)
            fprintf(f, "%.4f\t%.4f\t", frequency, frequency2);

         if (minMaxFrequencies)
            fprintf(f, "%.4f\t%.4f\t", minFrequency, maxFrequency);

         fprintf(f, "%.2f\t%.3f\t%.4g\t%s\n",
                  useStandardErrors ? statistics[marker] / weights[marker] : weights[marker],
                  useStandardErrors ? sqrt(1.0 / weights[marker]) : statistic,
                  pvalue,
                  (const char *) direction);
         count++;

         if (pvalue < smallp)
            {
            smallp = pvalue;
            smallm = markerLookup[i];
            }
         }

   printf("## Completed meta-analysis for %d markers!\n", count);

   if (smallp < 1.0)
      printf("## Smallest p-value is %.4g at marker '%s'\n", smallp, (const char *) smallm);

   if (skipped)
      printf("## A total of %d markers with less than MINWEIGHT = %g were skipped\n", skipped, minweight);

   fclose(f);

   f = fopen(filename + ".info", "wt");

   fprintf(f, "# This file contains a short description of the columns in the\n"
              "# meta-analysis summary file, named '%s'\n\n"
              "# Marker    - this is the marker name\n"
              "# Allele1   - the first allele for this marker in the first file where it occurs\n"
              "# Allele2   - the second allele for this marker in the first file where it occurs\n"
              "# Weight    - the sum of the individual study weights (typically, N) for this marker\n"
              "# Z-score   - the combined z-statistic for this marker\n"
              "# Direction - summary of effect direction for each study, with one '+' or '-' per study\n\n"
              "# Input for this meta-analysis was stored in the files:\n",
              (const char *) filename);

   for (int i = 0; i < filenames.Length(); i++)
      fprintf(f, "# --> Input File %d : %s\n", i + 1, (const char *) filenames[i]);
   fprintf(f, "\n");

   fclose(f);
   }

bool flip = false;

void ProcessFile(String & filename)
   {
   IFILE f = ifopen(filename, "rb");

   if (f == NULL)
      {
      printf("## Failed to open file '%s'\n", (const char *) filename);
      return;
      }

   printf("## Processing file '%s'\n", (const char *) filename);

   String input;
   StringArray tokens;

   tokens.ReplaceTokens(input.ReadLine(f), separators);

   if (tokens.Length() == 0)
      {
      printf("## ERROR: Header line is empty\n");
      ifclose(f);
      return;
      }

   int markerColumn = tokens.SlowFind(markerLabel);
   if (markerColumn < 0)
      {
      printf("## ERROR: No '%s' column found\n", (const char *) markerLabel);
      ifclose(f);
      return;
      }

   int pvalueColumn = tokens.SlowFind(pvalueLabel);
   if (pvalueColumn < 0 && !useStandardErrors)
      {
      printf("## ERROR: No '%s' column found\n", (const char *) pvalueLabel);
      ifclose(f);
      return;
      }

   int effectColumn = tokens.SlowFind(effectLabel);
   if (effectColumn < 0)
      if (!useStandardErrors)
         printf("## WARNING: No '%s' column found -- assuming all effects are positive!\n", (const char *) effectLabel);
      else
         {
         printf("## ERROR: No '%s' column found\n", (const char *) effectLabel);
         ifclose(f);
         return;
         }

   int weightColumn = tokens.SlowFind(weightLabel);
   if (weightColumn < 0 && !useStandardErrors)
      printf("## WARNING: No '%s' column found -- using DEFAULTWEIGHT = %g\n", (const char *) weightLabel, weight);

   int firstColumn = tokens.SlowFind(firstAllele);
   int secondColumn = tokens.SlowFind(secondAllele);
   bool guessAlleles = false;

   if (firstColumn >= 0 && secondColumn < 0)
      {
      printf("## WARNING: Column '%s' not found, will try to guess second allele\n",
             (const char *) secondAllele);
      guessAlleles = true;
      }

   if (firstColumn < 0)
      printf("## WARNING: Column '%s', '%s' or both not found -- assuming all effects refer to same allele\n",
             (const char *) firstAllele, (const char *) secondAllele, firstColumn = 0);

   int stderrColumn = tokens.SlowFind(stderrLabel);
   if (stderrColumn < 0 && useStandardErrors)
      {
      printf("## ERROR: Analysis based on standard errors requested but no '%s' column found\n", (const char *) stderrLabel);
      ifclose(f);
      return;
      }

   bool useFrequencies = minMaxFrequencies || averageFrequencies;
   int freqColumn = tokens.SlowFind(frequencyLabel);
   if (freqColumn < 0 && useFrequencies)
      {
      printf("## ERROR: Averaging of allele frequencies requested, but no '%s' column found\n", (const char *) frequencyLabel);
      ifclose(f);
      return;
      }

   int strandColumn = tokens.SlowFind(strandLabel);
   if (strandColumn < 0 && useStrand)
      {
      printf("## ERROR: Strand column labeleled '%s' not found\n", (const char *) strandLabel);
      ifclose(f);
      return;
      }

   SetupFilters(tokens);

   int minColumns = markerColumn + 1;
   if (weightColumn >= minColumns && !useStandardErrors) minColumns = weightColumn + 1;
   if (effectColumn >= minColumns) minColumns = effectColumn + 1;
   if (pvalueColumn >= minColumns && !useStandardErrors) minColumns = pvalueColumn + 1;
   if (firstColumn >= minColumns)  minColumns = firstColumn + 1;
   if (secondColumn >= minColumns) minColumns = secondColumn + 1;
   if (stderrColumn >= minColumns && useStandardErrors) minColumns = stderrColumn + 1;
   if (strandColumn >= minColumns && useStrand) minColumns = strandColumn + 1;
   if (freqColumn >= minColumns && useFrequencies) minColumns = freqColumn + 1;
   if (filterColumn.Max() >= minColumns) minColumns = filterColumn.Max() + 1;

   Vector backupStatistics, backupWeights, backupFrequencies, backupFrequencies2;
   Vector chiSquareds;

   if (genomicControl)
      {
      backupStatistics = statistics;
      statistics.Zero();
      }

   if (genomicControl && useStandardErrors)
      {
      backupFrequencies2 = frequencies2;
      backupFrequencies = frequencies;
      backupWeights = weights;

      frequencies2.Zero();
      frequencies.Zero();
      weights.Zero();
      }

   String direction;
   direction.Fill('?', allele1.Length());

   int invalid = 0;
   int badAlleles = 0;
   int badGuesses = 0;
   int duplicates = 0;
   int blindGuesses = 0;

   filenames.Push(filename);

   while (!ifeof(f))
      {
      if (separators.Length() != 1)
         tokens.ReplaceTokens(input.ReadLine(f).Trim(), separators);
      else
         tokens.ReplaceColumns(input.ReadLine(f), separators[0]);

      if (input[0] == '#') continue;
      if (tokens.Length() < minColumns) continue;

      if (!ApplyFilter(tokens)) continue;

      int marker = markerLookup.Integer(tokens[markerColumn]);

      if (marker < 0 && guessAlleles)
         {
         if (++blindGuesses > maxWarnings) continue;
         printf("## WARNING: Can't guess alleles for '%s', marker not seen in previous files\n",
                (const char *) tokens[markerColumn]);
         continue;
         }

      if (marker < 0)
         {
         markerLookup.SetInteger(tokens[markerColumn], marker = markerLookup.Entries());
         statistics.Push(0);
         if (averageFrequencies) frequencies.Push(0);
         if (averageFrequencies) frequencies2.Push(0);
         if (minMaxFrequencies) minFrequencies.Push(_NAN_);
         if (minMaxFrequencies) maxFrequencies.Push(_NAN_);
         weights.Push(0);
         allele1.Push("");
         allele2.Push("");
         original_flipped += '?';
         direction += '?';
         }

      double w, z;

      if (!useStandardErrors)
         {
         double p = tokens[pvalueColumn].AsDouble();

         if (p <= 0 || p > 1.0)
            {
            if (++invalid > maxWarnings) continue;
            printf("## WARNING: Invalid pvalue for marker %s, ignored\n", (const char *) tokens[markerColumn]);
            continue;
            }

         z = -ninv(tokens[pvalueColumn].AsDouble() * 0.5);
         w = weightColumn >= 0 ? tokens[weightColumn].AsDouble() : weight;

         if (effectColumn >= 0 && tokens[effectColumn][0] == '-')
            z *= -1;
         }
      else
         {
         double eff = tokens[effectColumn].AsDouble();
         double sd  = tokens[stderrColumn].AsDouble();

         if (sd <= 0)
            {
            if (++invalid > maxWarnings) continue;
            printf("## WARNING: Invalid standard error for marker %s, ignored\n", (const char *) tokens[markerColumn]);
            continue;
            }

         z = eff;
         w = 1.0 / (sd * sd);
         }

      double freq = 0.0;
      if (useFrequencies)
         freq = tokens[freqColumn].AsDouble();

      if (flip)
         z *= -1;

      if (firstColumn >= 0 && secondColumn >= 0)
         {
         NumbersToLetters(tokens[firstColumn]);
         NumbersToLetters(tokens[secondColumn]);

         if (useStrand && tokens[strandColumn] == "-")
            {
            FlipAllele(tokens[firstColumn]);
            FlipAllele(tokens[secondColumn]);
            }

         bool flip = FlipAlleles(tokens[firstColumn], tokens[secondColumn], z, freq);

         if (allele1[marker] == "")
            allele1[marker] = tokens[firstColumn],
            allele2[marker] = tokens[secondColumn],
            original_flipped[marker] = flip ? 'Y' : 'N';
         else
            if (tokens[firstColumn] != allele1[marker] ||
                tokens[secondColumn] != allele2[marker])
                {
                if (++badAlleles > maxWarnings) continue;
                printf("## WARNING: Bad alleles for marker '%s', expecting '%s/%s' found '%s/%s'\n",
                         (const char *) tokens[markerColumn],
                         (const char *) allele1[marker],
                         (const char *) allele2[marker],
                         (const char *) tokens[firstColumn],
                         (const char *) tokens[secondColumn]);
                continue;
                }
         }
      else if (guessAlleles)
         {
         tokens.Push("");
         NumbersToLetters(tokens[firstColumn]);
         if (!GuessSecondAllele(marker, tokens[firstColumn], tokens.Last()))
             {
             if (++badGuesses <= maxWarnings)
                printf("## WARNING: Bad allele or ambiguous strand for marker '%s', expecting '%s/%s' found '%s'\n",
                            (const char *) tokens[markerColumn],
                            (const char *) allele1[marker],
                            (const char *) allele2[marker],
                            (const char *) tokens[firstColumn]);
             continue;
             }
         FlipAlleles(tokens[firstColumn], tokens.Last(), z, freq);
         }

      if (direction[marker] != '?')
         {
         if (++duplicates > maxWarnings) continue;
         printf("## WARNING: Marker '%s' duplicated in input, first instance used, others skipped\n",
                (const char *) tokens[markerColumn]);
         continue;
         }

      direction[marker] = z == 0.0 ? '0' : (z > 0.0 ? '+' : '-');

      if (!useStandardErrors)
         {
         statistics[marker] += sqrt(w) * z;
         weights[marker] += w;
         chiSquareds.Push(z * z);
         }
      else
         {
         statistics[marker] += w * z;
         weights[marker] += w;
         chiSquareds.Push(z * z * w);
         }

      if (averageFrequencies)
         {
         frequencies[marker] += w * freq;
         frequencies2[marker] += w * freq * freq;
         }

      if (minMaxFrequencies)
         {
         if (minFrequencies[marker] > freq || minFrequencies[marker] == _NAN_)
            minFrequencies[marker] = freq;
         if (maxFrequencies[marker] < freq || maxFrequencies[marker] == _NAN_)
            maxFrequencies[marker] = freq;
         }
      }

   if (invalid > maxWarnings)
      printf("## WARNING: Invalid %s for %d other markers also ignored\n",
             useStandardErrors ? "standard errors" : "p-values", invalid - maxWarnings);

   if (badGuesses > maxWarnings)
      printf("## WARNING: Failed to guess second allele for %d other markers\n", badGuesses - maxWarnings);

   if (duplicates > maxWarnings)
      printf("## WARNING: An additional %d rows with duplicate marker names were ignored\n", duplicates - maxWarnings);

   if (badAlleles > maxWarnings)
      printf("## WARNING: Allele names don't match previous occurences at %d additional markers\n", badAlleles - maxWarnings);

   if (blindGuesses > maxWarnings)
      printf("## WARNING: An additional %d markers whose alleles couldn't be guessed were skipped\n", blindGuesses - maxWarnings);

   FilterSummary();

   directions.Push(direction);

   if (genomicControl)
      {
      if (chiSquareds.Length() == 0)
         printf("## WARNING: Genomic control parameter cannot be calculated, no valid input\n");
      else
         {
         chiSquareds.Sort();

         double gc = chiSquareds[0.5] / 0.4549364;

         if (gc <= 1.0)
            printf("## Genomic control parameter is %.3f, no adjustment made\n", gc);
         else
            printf("## Genomic control parameter is %.3f, adjusting test statistics\n", gc);

         if (!useStandardErrors)
            statistics *= 1.0 / sqrt(gc);
         else
            statistics *= 1.0 / gc;

         for (int i  = 0; i < backupStatistics.Length(); i++)
            statistics[i] += backupStatistics[i];

         if (useStandardErrors)
            {
            weights *= 1.0 / gc;

            if (averageFrequencies) frequencies *= 1.0 / gc;
            if (averageFrequencies) frequencies2 *= 1.0 / gc;

            for (int i = 0; i < backupWeights.Length(); i++)
               weights[i] += backupWeights[i];

            for (int i = 0; i < backupFrequencies.Length(); i++)
               {
               frequencies2[i] += backupFrequencies2[i];
               frequencies[i] += backupFrequencies[i];
               }
            }
         }
      }

   ifclose(f);
   }

void ShowHelp(bool startup = false)
   {
   const char * setting = startup ? "default" : "setting";

   printf("# This program faciliates meta-analysis of genome-wide association studies.\n"
          "# Commonly used commands are listed below:\n"
          "#\n"
          "# Options for describing input files ...\n"
          "#   SEPARATOR        [WHITESPACE|COMMA|BOTH|TAB] (default = WHITESPACE)\n"
          "#   MARKERLABEL      [LABEL]                     (%s = '%s')\n"
          "#   ALLELELABELS     [LABEL1 LABEL2]             (%s = '%s','%s')\n"
          "#   EFFECTLABEL      [LABEL]                     (%s = '%s')\n"
          "#   FLIP\n"
          "#\n"
          "# Options for filtering input files ...\n"
          "#   ADDFILTER        [LABEL CONDITION VALUE]     (example = ADDFILTER N > 10)\n"
          "#   REMOVEFILTERS\n"
          "#\n"
          "# Options for sample size weighted meta-analysis ...\n"
          "#   WEIGHTLABEL      [LABEL]                     (%s = '%s')\n"
          "#   PVALUELABEL      [LABEL]                     (%s = '%s')\n"
          "#   DEFAULTWEIGHT    [NUMBER]                    (%s = %.1f)\n"
          "#   MINWEIGHT        [NUMBER]                    (%s = %.1f)\n"
          "#\n"
          "# Options for inverse variance weighted meta-analysis ...\n"
          "#   STDERRLABEL      [LABEL]                     (%s = '%s')\n"
          "#   SCHEME           [SAMPLESIZE|STDERR]         (%s = %s)\n"
          "#\n"
          "# Options to enable tracking of allele frequencies ...\n"
          "#   AVERAGEFREQ      [ON|OFF]                    (%s = %s)\n"
          "#   MINMAXFREQ       [ON|OFF]                    (%s = %s)\n"
          "#   FREQLABEL        [LABEL]                     (%s = '%s')\n"
          "#\n"
          "# Options to enable explicit strand information ...\n"
          "#   USESTRAND        [ON|OFF]                    (%s = %s)\n"
          "#   STRANDLABEL      [LABEL]                     (%s = '%s')\n"
          "#\n"
          "# Automatic genomic control correction of input statistics ...\n"
          "#   GENOMICCONTROL   [ON|OFF]                    (%s = %s)\n"
          "#\n"
          "# Options for general run control ...\n"
          "#   PROCESSFILE      [FILENAME]\n"
          "#   OUTFILE          [PREFIX SUFFIX]             (default = 'METAANALYSIS','.TBL')\n"
          "#   MAXWARNINGS      [NUMBER]                    (%s = %d)\n"
          "#   ANALYZE\n"
          "#   CLEAR\n\n",
          setting, (const char *) markerLabel,
          setting, (const char *) firstAllele, (const char *) secondAllele,
          setting, (const char *) effectLabel,
          setting, (const char *) weightLabel,
          setting, (const char *) pvalueLabel,
          setting, weight,
          setting, minweight,
          setting, (const char *) stderrLabel,
          setting, useStandardErrors ? "STDERR" : "SAMPLESIZE",
          setting, averageFrequencies ? "ON" : "OFF",
          setting, minMaxFrequencies ? "ON" : "OFF",
          setting, (const char *) frequencyLabel,
          setting, useStrand ? "ON" : "OFF",
          setting, (const char *) strandLabel,
          setting, genomicControl ? "ON" : "OFF",
          setting, maxWarnings);
   }

int main(int argc, char ** argv)
   {
   SetupCrashHandlers();

   String::caseSensitive = false;

   printf("MetaAnalysis Helper - (c) 2007 - 2008 Goncalo Abecasis\n\n");

   ShowHelp(true);

   String input;
   StringArray tokens;

   while (!feof(stdin))
      {
      input.ReadLine().Trim();
      tokens.ReplaceTokens(input);

      if (tokens.Length() == 0)
         continue;

      if (input[0] == '#') continue;

      if ((tokens[0].MatchesBeginningOf("ANALYZE") == 0 ||
          tokens[0].MatchesBeginningOf("ANALYSE") == 0) &&
          tokens[0].Length() > 1)
         {
         Analyze();
         continue;
         }

      if (tokens[0].MatchesBeginningOf("CLEAR") == 0)
         {
         ClearAll();
         continue;
         }

      if (tokens[0].MatchesBeginningOf("REMOVEFILTERS") == 0)
         {
         ClearFilters();
         continue;
         }

      if (tokens[0].MatchesBeginningOf("FLIP") == 0)
         {
         flip = !flip;
         printf("## All effects will %s flipped\n", flip ? "be" : "not be");
         continue;
         }

      if (tokens[0].MatchesBeginningOf("HELP") == 0)
         {
         ShowHelp();
         continue;
         }

      if (tokens[0].MatchesBeginningOf("QUIT") == 0 ||
          tokens[0].MatchesBeginningOf("EXIT") == 0 && tokens[0].Length() > 1)
         break;

      if (tokens.Length() == 1)
         {
         printf("##  Error processing your request ...\n");
         continue;
         }

      if (tokens[0].MatchesBeginningOf("DEFAULTWEIGHT") == 0)
         {
         weight = tokens[1].AsDouble();
         printf("## Set default weight to %.2f ...\n", weight);
         continue;
         }

      if (tokens[0].MatchesBeginningOf("WEIGHTLABEL") == 0)
         {
         weightLabel = tokens[1];
         printf("## Set weight header to %s ...\n", (const char *) weightLabel);
         continue;
         }

      if (tokens[0].MatchesBeginningOf("GENOMICCONTROL") == 0)
         {
         if (tokens[1] == "ON")
            {
            genomicControl = true;
            printf("## Genomic control correction of input statistics enabled\n");
            continue;
            }
         else if (tokens[1].MatchesBeginningOf("OFF") == 0 && tokens[1].Length() > 1)
            {
            genomicControl = false;
            printf("## Genomic control correction of input statistics disabled\n");
            continue;
            }
         }

      if (tokens[0].MatchesBeginningOf("USESTRAND") == 0 && tokens[0].Length() > 3)
         {
         if (tokens[1] == "ON")
            {
            useStrand = true;
            printf("## Strand flipping according to strand column enabled\n");
            continue;
            }
         else if (tokens[1].MatchesBeginningOf("OFF") == 0 && tokens[1].Length() > 1)
            {
            useStrand = false;
            printf("## Strand flipping according to strand column disabled\n");
            continue;
            }
         }

      if (tokens[0].MatchesBeginningOf("AVERAGEFREQUENCY") == 0 && tokens[0].Length() > 1)
         {
         if (markerLookup.Entries())
            {
            printf("## Meta-analysis in progress - before issuing this command, use CLEAR command\n");
            continue;
            }

         if (tokens[1] == "ON")
            {
            averageFrequencies = true;
            printf("## Averaging of allele frequencies enabled\n");
            continue;
            }
         else if (tokens[1].MatchesBeginningOf("OFF") == 0 && tokens[1].Length() > 1)
            {
            averageFrequencies = false;
            printf("## Average of allele frequencies disabled\n");
            continue;
            }
         }

      if (tokens[0].MatchesBeginningOf("MINMAXFREQUENCY") == 0 && tokens[0].Length() > 3)
         {
         if (markerLookup.Entries())
            {
            printf("## Meta-analysis in progress - before issuing this command, use CLEAR command\n");
            continue;
            }

         if (tokens[1] == "ON")
            {
            minMaxFrequencies = true;
            printf("## Tracking of extreme allele frequencies enabled\n");
            continue;
            }
         else if (tokens[1].MatchesBeginningOf("OFF") == 0 && tokens[1].Length() > 1)
            {
            minMaxFrequencies = false;
            printf("## Tracking of extreme allele frequencies disabled\n");
            continue;
            }
         }

      if (tokens[0].MatchesBeginningOf("MARKERLABEL") == 0 &&
          tokens[0].Length() > 2)
         {
         markerLabel = tokens[1];
         printf("## Set marker header to %s ...\n", (const char *) markerLabel);
         continue;
         }

      if (tokens[0].MatchesBeginningOf("MAXWARNINGS") == 0 &&
          tokens[0].Length() > 2)
          {
          maxWarnings = tokens[1].AsInteger();
          printf("## Set the maximum number of warnings per file to %d ...\n", maxWarnings);
          continue;
          }

      if (tokens[0].MatchesBeginningOf("PVALUELABEL") == 0 && tokens[0].Length() > 1)
         {
         pvalueLabel = tokens[1];
         printf("## Set p-value header to %s ...\n", (const char *) pvalueLabel);
         continue;
         }

      if (tokens[0].MatchesBeginningOf("EFFECTLABEL") == 0 && tokens[0].Length() > 1)
         {
         effectLabel = tokens[1];
         printf("## Set effect header to %s ...\n", (const char *) effectLabel);
         continue;
         }

      if (tokens[0].MatchesBeginningOf("FREQLABEL") == 0)
         {
         frequencyLabel = tokens[1];
         printf("## Set frequency header to %s ...\n", (const char *) frequencyLabel);
         if (!averageFrequencies) printf("## If you want frequencies to averaged, issue the 'AVERAGEFREQ ON' command\n");
         continue;
         }

      if (tokens[0].MatchesBeginningOf("STDERRLABEL") == 0 && tokens[0].Length() > 1)
         {
         stderrLabel = tokens[1];
         printf("## Set standard error header to %s ...\n", (const char *) stderrLabel);
         continue;
         }

      if (tokens[0].MatchesBeginningOf("STRANDLABEL") == 0 && tokens[0].Length() > 2)
         {
         strandLabel = tokens[1];
         printf("## Set strand header to %s ...\n", (const char *) strandLabel);
         if (!useStrand) printf("## If you want strand information to be used, issue the 'USESTRAND ON' command\n");
         continue;
         }

      if (tokens[0].MatchesBeginningOf("SCHEME") == 0 &&
          tokens[0].Length() > 1)
         if (markerLookup.Entries())
            {
            printf("## Meta-analysis in progress - before changing analysis scheme, use CLEAR command\n");
            continue;
            }
         else if (tokens[1].MatchesBeginningOf("SAMPLESIZE") == 0 &&
                  tokens[1].Length() > 1)
            {
            useStandardErrors = false;
            printf("## Meta-analysis will be based on sample sizes, p-values and direction of effect ...\n");
            continue;
            }
         else if (tokens[1].MatchesBeginningOf("STDERR") == 0 &&
                  tokens[1].Length() > 1)
            {
            useStandardErrors = true;
            printf("## Meta-analysis will be based on effect sizes and their standard errors ...\n");
            continue;
            }

      if (tokens[0].MatchesBeginningOf("SEPARATOR") == 0 &&
          tokens[0].Length() > 1)
         if (tokens[1].MatchesBeginningOf("WHITESPACE") == 0)
            {
            separators = " \t";
            printf("## Set column separator to WHITESPACE ...\n");
            continue;
            }
         else if (tokens[1].MatchesBeginningOf("COMMAS") == 0)
            {
            separators = ",";
            printf("## Set column separator to COMMAS ...\n");
            continue;
            }
         else if (tokens[1].MatchesBeginningOf("TABS") == 0)
            {
            separators = "\t";
            printf("## Set column separator to TAB ...\n");
            continue;
            }
         else if (tokens[1].MatchesBeginningOf("BOTH") == 0)
            {
            separators = " \t,";
            printf("## Set column separator to BOTH ...\n");
            continue;
            }

      if (tokens[0].MatchesBeginningOf("PROCESSFILE") == 0 && tokens[0].Length() > 1)
         {
         ProcessFile(tokens[1]);
         continue;
         }

      if (tokens[0].MatchesBeginningOf("MINWEIGHT") == 0 &&
          tokens[0].Length() > 3)
          {
          minweight = tokens[1].AsDouble();
          printf("## Set minimum weight for meta-analysis to %.2f ...\n", minweight);
          continue;
          }

      if (tokens.Length() == 2)
         {
         printf("##  Error processing your request ...\n");
         continue;
         }

      if (tokens[0].MatchesBeginningOf("ALLELELABELS") == 0 &&
          tokens[0].Length() > 1)
         {
         firstAllele = tokens[1];
         secondAllele = tokens[2];
         printf("## Set allele headers to %s and %s ...\n",
               (const char *) firstAllele, (const char *) secondAllele);
         continue;
         }

      if (tokens[0].MatchesBeginningOf("OUTFILE") == 0)
         {
         outfileround = 1;
         outfile = tokens[1] + "%d" + tokens[2];
         printf("## Set output file prefix and suffix to %s and %s ...\n",
                 (const char *) tokens[1], (const char *) tokens[2]);
         continue;
         }

      if (tokens[0].MatchesBeginningOf("ADDFILTER") == 0 &&
          tokens[0].Length() > 2)
         {
         AddFilter(tokens);
         continue;
         }

      printf("## Command not recognized - type HELP to list available commands\n");
      }
   }
 
